/**
 * @file
 * JS to convert the form browser validation messages into accessible ones.
 */
(function (Drupal, once, drupalSettings) {
  Drupal.behaviors.accessibleFormValidation = {
    attach(context) {
      const settings = drupalSettings.accessibleFormValidation;
      const errorMessageContainerClasses =
        settings && settings.errorMessageSelector
          ? settings.errorMessageSelector
          : 'form-item--error-message';
      const errorMessageContainerSelector = `.${errorMessageContainerClasses}`;
      const errorMessageContainerErrorClasses = 'invalid-feedback';
      const formItemSelector = '.form-item';
      const formItemErrorClasses = ['form-item--error', 'was-validated'];
      const formItemValidClasses = ['was-validated'];
      const labelErrorClasses = 'has-error';
      const inputInvalidClasses = ['is-invalid', 'error'];
      const inputValidClasses = ['is-valid'];
      const choicesJsInputClass = 'webform-choices';
      const choicesJsContainerClasses = 'choices';

      /**
       * Gets or creates an error message container for the input.
       *
       * @param {HTMLElement} input - The form input element.
       * @returns {HTMLElement} The error message container element.
       */
      const getErrorMessageContainer = (input) => {
        const container = input
          .closest(formItemSelector)
          ?.querySelector(errorMessageContainerSelector);
        if (container) {
          return container;
        }
        const errorMessageContainer = document.createElement('div');
        errorMessageContainer.classList.add(errorMessageContainerClasses);

        const inputContainer = input.closest(formItemSelector);
        if (inputContainer.querySelector('.form-item__description')) {
          // If there is a description, insert the error message container before it.
          inputContainer
            .querySelector('.form-item__description')
            .insertAdjacentElement('beforebegin', errorMessageContainer);
        } else {
          // Otherwise, append it to the end of the form item.
          inputContainer.appendChild(errorMessageContainer);
        }

        return errorMessageContainer;
      };

      /**
       * Marks an input as invalid and displays the validation message.
       *
       * @param {HTMLElement} input - The form input element to mark as invalid.
       */
      const setInputAsInvalid = (input) => {
        const inputContainer = input.closest(formItemSelector);
        const errorMessageContainer = getErrorMessageContainer(input);
        errorMessageContainer.innerText = input.validationMessage;
        errorMessageContainer.classList.add(errorMessageContainerErrorClasses);

        input.setAttribute('aria-invalid', 'true');
        input.classList.remove(...inputValidClasses);
        input.classList.add(...inputInvalidClasses);
        inputContainer.classList.remove(...formItemValidClasses);
        inputContainer.classList.add(...formItemErrorClasses);
        inputContainer.querySelector('label').classList.add(labelErrorClasses);
      };

      /**
       * Marks an input as valid and removes any error messages.
       *
       * @param {HTMLElement} input - The form input element to mark as valid.
       */
      const setInputAsValid = (input) => {
        const inputContainer = input.closest(formItemSelector);
        if (inputContainer.querySelector(errorMessageContainerSelector)) {
          inputContainer.querySelector(errorMessageContainerSelector).remove();
        }
        input.setAttribute('aria-invalid', 'false');
        input.classList.remove(...inputInvalidClasses);
        input.classList.add(...inputValidClasses);
        inputContainer.classList.remove(...formItemErrorClasses);
        inputContainer.classList.add(...formItemValidClasses);
        inputContainer
          .querySelector('label')
          .classList.remove(labelErrorClasses);
      };

      /**
       * Updates the validation UI state for a given input element.
       *
       * @param {HTMLElement} element - The input element to update the UI state for.
       */
      const updateValidationStateForInput = (element) => {
        // Check if the input is valid using the Constraint Validation API.
        const isInputValid = element.checkValidity();

        if (isInputValid) {
          setInputAsValid(element);
        } else {
          setInputAsInvalid(element);
        }
      };

      /**
       * Initializes the validation for a form field.
       *
       * @param {HTMLElement} input - The form input element to initialize.
       */
      const initField = (input) => {
        // Validate the input fields on change and blur.
        input.addEventListener('input', (event) => {
          updateValidationStateForInput(event.target);
        });

        input.addEventListener('blur', (event) => {
          updateValidationStateForInput(event.target);
        });

        // Choices.js support.
        if (input.classList.contains(choicesJsInputClass)) {
          input.addEventListener('change', (event) => {
            updateValidationStateForInput(event.target);
          });
          input
            .closest(formItemSelector)
            .addEventListener('focusout', (event) => {
              if (event.target.classList.contains(choicesJsContainerClasses)) {
                updateValidationStateForInput(input);
              }
            });
        }

        // Should be set to "false" before any validation occurs.
        input.setAttribute('aria-invalid', 'false');
      };

      /**
       * Handler for form submit event.
       *
       * @param {SubmitEvent} event - The form submission event.
       */
      const onSubmit = (event) => {
        const form = event.target;
        const formInputs = form.querySelectorAll('*:required');

        // Update the validation UI state for all inputs.
        formInputs.forEach(updateValidationStateForInput);

        const isFormValid = form.checkValidity();

        // Prevent form submission if any of the validation checks fail.
        if (!isFormValid) {
          event.preventDefault();
        }

        // Set the focus to the first invalid input.
        const firstInvalidInputEl = form.querySelector('input:invalid');
        // @TODO: Add compatibility with field groups module
        firstInvalidInputEl?.focus();
      };

      once('accessibleFormValidation', 'form', context).forEach((form) => {
        const formInputs = form.querySelectorAll('*:required');

        if (formInputs.length === 0) {
          return;
        }

        // Avoid default HTML5 validation.
        form.setAttribute('novalidate', 'novalidate');
        // Handle form submits validation via JS instead.
        form.addEventListener('submit', onSubmit);
      });

      once(
        'accessibleFormValidationInput',
        'input, textarea, select',
        context,
      ).forEach((input) => {
        if (input.required) {
          initField(input);
        }
      });
    },
  };
})(Drupal, once, drupalSettings);
